@@ -1,17 +1,18 @@
-use crate::parser::{Program, Op, OpType};
+use std::io::{ErrorKind, Read, Write};
+use std::mem;
+use std::process::exit;
 
 use cranelift::prelude::*;
 use cranelift_codegen::binemit::{NullStackMapSink, NullTrapSink};
+use cranelift_codegen::ir::FuncRef;
 use cranelift_codegen::settings::{self, Configurable};
 use cranelift_jit::{JITBuilder, JITModule};
-use cranelift_module::{default_libcall_names, Linkage, Module, FuncId};
-use std::mem;
-use std::io::{Read, Write, ErrorKind};
-use cranelift_codegen::ir::FuncRef;
-use std::cmp::Ordering;
+use cranelift_module::{default_libcall_names, FuncId, Linkage, Module};
+
 use crate::errors::CompilerError;
-use std::process::exit;
+use crate::ir::ops::{Op, OpType, LoopDecrement};
 use crate::OptimizeConfig;
+use crate::parser::Program;
 
 struct Builder<'a> {
     pointer_type: Type,
@@ -26,6 +27,9 @@ impl<'a> Builder<'a> {
     pub fn append_ops(&mut self, ops: &[Op]) {
         for op in ops {
             match &op.op_type {
+                OpType::Start => {
+                    // ignore
+                }
                 OpType::IncPtr(value) => self.inc_ptr(*value),
                 OpType::DecPtr(value) => self.dec_ptr(*value),
                 OpType::Inc(offset, value) => self.inc(*offset, *value),
@@ -39,16 +43,19 @@ impl<'a> Builder<'a> {
                 OpType::NzSub(src_offset, dest_offset, multi) => self.nz_sub(*src_offset, *dest_offset, *multi),
                 OpType::CSub(src_offset, dest_offset, value) => self.c_sub(*src_offset, *dest_offset, *value),
                 OpType::NzCSub(src_offset, dest_offset, value) => self.nz_c_sub(*src_offset, *dest_offset, *value),
+                OpType::Mul(src_offset, dest_offset, multi) => self.mul(*src_offset, *dest_offset, *multi),
+                OpType::NzMul(src_offset, dest_offset, multi) => self.nz_mul(*src_offset, *dest_offset, *multi),
                 OpType::Move(src_offset, dest_offset) => self._move(*src_offset, *dest_offset),
                 OpType::Copy(src_offset, dest_offset) => self.copy(*src_offset, *dest_offset),
-                OpType::DLoop(ops) => self.d_loop(ops),
-                OpType::LLoop(ops, offset) => self.l_loop(ops, *offset),
-                OpType::ILoop(ops, offset, step) => self.i_loop(ops, *offset, *step),
-                OpType::CLoop(ops, offset, iterations) => self.c_loop(ops, *offset, *iterations),
-                OpType::TNz(ops, offset) => self.tnz(ops, *offset),
+                OpType::DLoop(ops, _) => self.d_loop(ops),
+                OpType::LLoop(ops, _) => self.l_loop(ops),
+                OpType::ILoop(ops, step, decrement, _) => self.i_loop(ops, *step, *decrement),
+                OpType::CLoop(ops, iterations, decrement, _) => self.c_loop(ops, *iterations, *decrement),
+                OpType::TNz(ops, _) => self.tnz(ops),
                 OpType::SearchZero(step) => self.search_zero(*step),
-                OpType::PutChar => self.put_char(),
-                OpType::GetChar => self.get_char(),
+                OpType::PutChar(offset) => self.put_char(*offset),
+                OpType::PutString(array) => self.put_string(array),
+                OpType::GetChar(offset) => self.get_char(*offset),
             }
         }
     }
@@ -57,24 +64,6 @@ impl<'a> Builder<'a> {
         self.bcx.ins().iconst(types::I8, value as i64)
     }
 
-    fn const_isize(&mut self, value: isize) -> Value {
-        self.bcx.ins().iconst(self.pointer_type, value as i64)
-    }
-
-    fn add_offset(&mut self, value: Value, offset: isize) -> Value {
-        match offset.cmp(&0) {
-            Ordering::Equal => value,
-            Ordering::Greater => {
-                let offset = self.const_isize(offset);
-                self.bcx.ins().iadd(value, offset)
-            }
-            Ordering::Less => {
-                let offset = self.const_isize(-offset);
-                self.bcx.ins().isub(value, offset)
-            }
-        }
-    }
-
     fn load(&mut self, offset: isize) -> Value {
         self.bcx.ins().load(types::I8, MemFlags::new(), self.heap_ptr, offset as i32)
     }
@@ -174,6 +163,20 @@ impl<'a> Builder<'a> {
         self.set(src_offset, 0);
     }
 
+    fn nz_mul(&mut self, src_offset: isize, dest_offset: isize, multi: u8) {
+        let source = self.load(src_offset);
+
+        let multi = self.const_u8(multi);
+        let target = self.bcx.ins().imul(source, multi);
+
+        self.store(dest_offset, target);
+    }
+
+    fn mul(&mut self, src_offset: isize, dest_offset: isize, multi: u8) {
+        self.nz_mul(src_offset, dest_offset, multi);
+        self.set(src_offset, 0);
+    }
+
     fn copy(&mut self, src_offset: isize, dest_offset: isize) {
         let value = self.load(src_offset);
         self.store(dest_offset, value);
@@ -184,17 +187,24 @@ impl<'a> Builder<'a> {
         self.set(src_offset, 0);
     }
 
-    fn get_char(&mut self) {
+    fn get_char(&mut self, offset: isize) {
         let results = self.bcx.ins().call(self.get_char_func, &[self.env]);
         let results = self.bcx.inst_results(results)[0];
-        self.store(0, results);
+        self.store(offset, results);
     }
 
-    fn put_char(&mut self) {
-        let value = self.load(0);
+    fn put_char(&mut self, offset: isize) {
+        let value = self.load(offset);
         self.bcx.ins().call(self.put_char_func, &[self.env, value]);
     }
 
+    fn put_string(&mut self, array: &[u8]) {
+        for value in array {
+            let value = self.const_u8(*value);
+            self.bcx.ins().call(self.put_char_func, &[self.env, value]);
+        }
+    }
+
     fn d_loop(&mut self, ops: &[Op]) {
         let head = self.bcx.create_block();
         self.bcx.append_block_param(head, self.pointer_type);
@@ -227,13 +237,12 @@ impl<'a> Builder<'a> {
         self.set(0, 0);
     }
 
-    fn l_loop(&mut self, ops: &[Op], offset: isize) {
+    fn l_loop(&mut self, ops: &[Op]) {
         let head = self.bcx.create_block();
         self.bcx.append_block_param(head, self.pointer_type);
 
         let body = self.bcx.create_block();
         self.bcx.append_block_param(body, self.pointer_type);
-        self.bcx.append_block_param(body, self.pointer_type);
 
         let next = self.bcx.create_block();
         self.bcx.append_block_param(next, self.pointer_type);
@@ -245,17 +254,16 @@ impl<'a> Builder<'a> {
         self.heap_ptr = self.bcx.block_params(head)[0];
 
         let init_heap_ptr = self.heap_ptr;
-        let loop_heap_ptr = self.add_offset(init_heap_ptr, offset);
 
         let value = self.load(0);
         self.bcx.ins().brz(value, next, &[init_heap_ptr]);
-        self.bcx.ins().fallthrough(body, &[loop_heap_ptr, init_heap_ptr]);
+        self.bcx.ins().fallthrough(body, &[init_heap_ptr]);
 
         // Loop Body
         self.bcx.switch_to_block(body);
-        self.heap_ptr = self.bcx.block_params(body)[0];
+        let heap_ptr = self.bcx.block_params(body)[0];
+        self.heap_ptr = heap_ptr;
         self.append_ops(ops);
-        let heap_ptr = self.bcx.block_params(body)[1];
         self.bcx.ins().jump(head, &[heap_ptr]);
 
         // Start next block after loop
@@ -266,14 +274,13 @@ impl<'a> Builder<'a> {
         self.set(0, 0);
     }
 
-    fn i_loop(&mut self, ops: &[Op], offset: isize, step: u8) {
+    fn i_loop(&mut self, ops: &[Op], step: u8, decrement: LoopDecrement) {
         let head = self.bcx.create_block();
         self.bcx.append_block_param(head, self.pointer_type);
         self.bcx.append_block_param(head, types::I8);
 
         let body = self.bcx.create_block();
         self.bcx.append_block_param(body, self.pointer_type);
-        self.bcx.append_block_param(body, self.pointer_type);
         self.bcx.append_block_param(body, types::I8);
 
         let next = self.bcx.create_block();
@@ -290,21 +297,33 @@ impl<'a> Builder<'a> {
         let counter = self.bcx.block_params(head)[1];
 
         let init_heap_ptr = self.heap_ptr;
-        let loop_heap_ptr = self.add_offset(init_heap_ptr, offset);
 
         self.bcx.ins().brz(counter, next, &[init_heap_ptr]);
-        self.bcx.ins().fallthrough(body, &[loop_heap_ptr, init_heap_ptr, counter]);
+        self.bcx.ins().fallthrough(body, &[init_heap_ptr, counter]);
 
         // Loop Body
         self.bcx.switch_to_block(body);
-        self.heap_ptr = self.bcx.block_params(body)[0];
-        let counter = self.bcx.block_params(body)[2];
+        let heap_ptr = self.bcx.block_params(body)[0];
+        self.heap_ptr = heap_ptr;
+        let mut counter = self.bcx.block_params(body)[1];
+        let step = self.const_u8(step);
+
+        if decrement == LoopDecrement::Pre {
+            counter = self.bcx.ins().isub(counter, step);
+            self.store(0, counter);
+        }
 
         self.append_ops(ops);
 
-        let step = self.const_u8(step);
-        let counter = self.bcx.ins().isub(counter, step);
-        let heap_ptr = self.bcx.block_params(body)[1];
+        self.heap_ptr = heap_ptr;
+
+        if decrement == LoopDecrement::Post {
+            counter = self.bcx.ins().isub(counter, step);
+            self.store(0, counter);
+        } else if decrement == LoopDecrement::Auto {
+            counter = self.bcx.ins().isub(counter, step);
+        }
+
         self.bcx.ins().jump(head, &[heap_ptr, counter]);
 
         // Start next block after loop
@@ -313,20 +332,24 @@ impl<'a> Builder<'a> {
         self.set(0, 0);
     }
 
-    fn c_loop(&mut self, ops: &[Op], offset: isize, iterations: u8) {
+    fn c_loop(&mut self, ops: &[Op], iterations: u8, decrement: LoopDecrement) {
         let head = self.bcx.create_block();
         self.bcx.append_block_param(head, self.pointer_type);
         self.bcx.append_block_param(head, types::I8);
 
         let body = self.bcx.create_block();
         self.bcx.append_block_param(body, self.pointer_type);
-        self.bcx.append_block_param(body, self.pointer_type);
         self.bcx.append_block_param(body, types::I8);
 
         let next = self.bcx.create_block();
         self.bcx.append_block_param(next, self.pointer_type);
 
         let iterations = self.const_u8(iterations);
+
+        if decrement == LoopDecrement::Post {
+            self.store(0, iterations);
+        }
+
         self.bcx.ins().fallthrough(head, &[self.heap_ptr, iterations]);
 
         // Head with condition
@@ -335,21 +358,34 @@ impl<'a> Builder<'a> {
         let counter = self.bcx.block_params(head)[1];
 
         let init_heap_ptr = self.heap_ptr;
-        let loop_heap_ptr = self.add_offset(init_heap_ptr, offset);
 
         self.bcx.ins().brz(counter, next, &[init_heap_ptr]);
-        self.bcx.ins().fallthrough(body, &[loop_heap_ptr, init_heap_ptr, counter]);
+        self.bcx.ins().fallthrough(body, &[init_heap_ptr, counter]);
 
         // Loop Body
         self.bcx.switch_to_block(body);
-        self.heap_ptr = self.bcx.block_params(body)[0];
-        let counter = self.bcx.block_params(body)[2];
+        let heap_ptr = self.bcx.block_params(body)[0];
+        self.heap_ptr = heap_ptr;
+        let mut counter = self.bcx.block_params(body)[1];
+
+        let step = self.const_u8(1);
+
+        if decrement == LoopDecrement::Pre {
+            counter = self.bcx.ins().isub(counter, step);
+            self.store(0, counter);
+        }
 
         self.append_ops(ops);
 
-        let step = self.const_u8(1);
-        let counter = self.bcx.ins().isub(counter, step);
-        let heap_ptr = self.bcx.block_params(body)[1];
+        self.heap_ptr = heap_ptr;
+
+        if decrement == LoopDecrement::Post {
+            counter = self.bcx.ins().isub(counter, step);
+            self.store(0, counter);
+        } else if decrement == LoopDecrement::Auto {
+            counter = self.bcx.ins().isub(counter, step);
+        }
+
         self.bcx.ins().jump(head, &[heap_ptr, counter]);
 
         // Start next block after loop
@@ -358,7 +394,7 @@ impl<'a> Builder<'a> {
         self.set(0, 0);
     }
 
-    fn tnz(&mut self, ops: &[Op], offset: isize) {
+    fn tnz(&mut self, ops: &[Op]) {
         let body = self.bcx.create_block();
         self.bcx.append_block_param(body, self.pointer_type);
 
@@ -374,7 +410,6 @@ impl<'a> Builder<'a> {
         self.heap_ptr = self.bcx.block_params(body)[0];
 
         let init_heap_ptr = self.heap_ptr;
-        self.heap_ptr = self.add_offset(init_heap_ptr, offset);
 
         self.append_ops(ops);
 
@@ -479,7 +514,6 @@ pub struct CompiledJitModule {
 }
 
 impl CompiledJitModule {
-
     /// Compile program
     pub fn new(program: &Program, opt_mode: &OptimizeConfig) -> Result<CompiledJitModule, CompilerError> {
         let mut flag_builder = settings::builder();
@@ -634,10 +668,14 @@ impl Drop for CompiledJitModule {
 
 #[cfg(test)]
 mod tests {
-    use crate::{parse, optimize, OptimizeConfig};
-    use std::io::{Cursor, Write, Read};
-    use crate::parser::{Program, Op};
+    use std::io::{Cursor, Read, Write};
+
+    use crate::{OptimizeConfig, parse, optimize_with_config};
+    use crate::ir::ops::{Op, LoopDecrement};
+    use crate::parser::Program;
+
     use super::CompiledJitModule;
+    use crate::ir::opt_info::BlockInfo;
 
     fn run<R: Read, W: Write>(program: &Program, input: R, output: W) -> Vec<u8> {
         CompiledJitModule::new(program, &OptimizeConfig::o2()).unwrap().execute(input, output)
@@ -646,7 +684,8 @@ mod tests {
     #[test]
     fn test_simple() {
         let mut program = parse("++>>++++-<-").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -661,7 +700,8 @@ mod tests {
     #[test]
     fn test_output() {
         let mut program = parse(",+.").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"a";
         let mut output = Vec::new();
@@ -678,7 +718,8 @@ mod tests {
                 >++++++++[<++++>-] <.>+++++++++++[<++++++++>-]<-.--------.+++
                 .------.--------.[-]>++++++++[<++++>- ]<+.[-]++++++++++."
         ).unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -692,7 +733,8 @@ mod tests {
     fn test_dloop() {
         let mut program = parse(">,<++[->+<]>."
         ).unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"1";
         let mut output = Vec::new();
@@ -705,7 +747,8 @@ mod tests {
     #[test]
     fn test_hello_world_v2() {
         let mut program = parse(include_str!("../../../test_programs/hello_world.bf")).unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -718,7 +761,8 @@ mod tests {
     #[test]
     fn test_count_loop() {
         let mut program = parse("++++[->++<]").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -732,7 +776,8 @@ mod tests {
     #[test]
     fn test_count_loop_inv() {
         let mut program = parse("++++[->++<]").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -746,7 +791,8 @@ mod tests {
     #[test]
     fn test_totally_empty() {
         let mut program = parse("").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -759,7 +805,8 @@ mod tests {
     #[test]
     fn test_optimized_empty() {
         let mut program = parse("[>+++++<-]").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -772,7 +819,8 @@ mod tests {
     #[test]
     fn test_multiply() {
         let mut program = parse(">>++<<++[>+++++<-]").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -786,7 +834,8 @@ mod tests {
     #[test]
     fn test_divide() {
         let mut program = parse(">>++<<++++++++++++[>+<---]").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -800,7 +849,8 @@ mod tests {
     #[test]
     fn test_pow_5_3() {
         let mut program = parse("+++++[>+++++[>+++++<-]<-]").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -815,7 +865,8 @@ mod tests {
     #[test]
     fn test_bad_count_loop() {
         let mut program = parse("++++++++[---->+<++]").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -829,7 +880,8 @@ mod tests {
     #[test]
     fn test_conditional_set() {
         let mut program = parse("+>>++<<[->[-]++++++<]").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -844,7 +896,8 @@ mod tests {
     #[test]
     fn test_overwrite_prev_set() {
         let mut program = parse("++++[-]++").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -857,7 +910,8 @@ mod tests {
     #[test]
     fn test_constant_loop() {
         let mut program = parse("+++++++++++++[>+>++>>+++<<<<-]->>>>>>+").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -876,7 +930,8 @@ mod tests {
     #[test]
     fn test_hello_fizz() {
         let mut program = parse(include_str!("../../../test_programs/fizz.bf")).unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -889,7 +944,8 @@ mod tests {
     #[test]
     fn test_hello_fizzbuzz() {
         let mut program = parse(include_str!("../../../test_programs/fizzbuzz.bf")).unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -902,7 +958,8 @@ mod tests {
     #[test]
     fn test_bottles() {
         let mut program = parse(include_str!("../../../test_programs/bottles.bf")).unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -915,7 +972,8 @@ mod tests {
     #[test]
     fn test_factor() {
         let mut program = parse(include_str!("../../../test_programs/factor.bf")).unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = include_bytes!("../../../test_programs/factor.bf.in");
         let mut output = Vec::new();
@@ -928,7 +986,8 @@ mod tests {
     #[test]
     fn test_life() {
         let mut program = parse(include_str!("../../../test_programs/life.bf")).unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = include_bytes!("../../../test_programs/life.bf.in");
         let mut output = Vec::new();
@@ -1154,6 +1213,43 @@ mod tests {
         assert_eq!(heap[3], 0u8.wrapping_sub(8));
     }
 
+    #[test]
+    fn test_mul() {
+        let program = Program {
+            ops: vec![
+                Op::set(0..1, 5),
+                Op::mul(1..2, 1, 2),
+                Op::mul(2..3, 3, 1)
+            ]
+        };
+
+        let input = b"";
+        let mut output = Vec::new();
+
+        let heap = run(&program, Cursor::new(input), &mut output);
+
+        assert_eq!(&heap[0..5], &[0, 10, 0, 0, 0]);
+    }
+
+    #[test]
+    fn test_nz_mul_mul() {
+        let program = Program {
+            ops: vec![
+                Op::set(0..1, 5),
+                Op::nz_mul(1..2, 1, 2),
+                Op::nz_mul(2..3, 3, 3),
+                Op::mul(2..3, 4, 4)
+            ]
+        };
+
+        let input = b"";
+        let mut output = Vec::new();
+
+        let heap = run(&program, Cursor::new(input), &mut output);
+
+        assert_eq!(&heap[0..6], &[0, 10, 0, 15, 20, 0]);
+    }
+
     #[test]
     fn test_move() {
         let program = Program {
@@ -1238,7 +1334,7 @@ mod tests {
                     Op::inc(1..2, 1),
                     Op::inc(1..2, 1),
                     Op::dec_ptr(1..2, 1),
-                ]),
+                ], BlockInfo::new_empty()),
             ]
         };
 
@@ -1257,12 +1353,13 @@ mod tests {
             ops: vec![
                 Op::set(0..1, 5),
                 Op::l_loop(1..4, vec![
+                    Op::inc_ptr(1..2, 1),
                     Op::dec_ptr(1..2, 1),
                     Op::dec(1..2, 1),
                     Op::inc_ptr(1..2, 2),
                     Op::inc(1..2, 1),
                     Op::inc(1..2, 1),
-                ], 1),
+                ], BlockInfo::new_empty()),
             ]
         };
 
@@ -1282,9 +1379,10 @@ mod tests {
             ops: vec![
                 Op::set(0..1, 5),
                 Op::i_loop(1..4, vec![
+                    Op::inc_ptr(1..2, 2),
                     Op::inc(1..2, 1),
                     Op::inc(1..2, 1),
-                ], 2, 1),
+                ], 1, BlockInfo::new_empty()),
             ]
         };
 
@@ -1304,9 +1402,10 @@ mod tests {
             ops: vec![
                 Op::set(0..1, 10),
                 Op::i_loop(1..4, vec![
+                    Op::inc_ptr(1..2, 2),
                     Op::inc(1..2, 1),
                     Op::inc(1..2, 1),
-                ], 2, 2),
+                ], 2, BlockInfo::new_empty()),
             ]
         };
 
@@ -1320,15 +1419,59 @@ mod tests {
         assert_eq!(heap[2], 10);
     }
 
+    #[test]
+    fn test_i_loop_pre() {
+        let program = Program {
+            ops: vec![
+                Op::set(0..1, 5),
+                Op::i_loop_with_decrement(1..4, vec![
+                    Op::inc_ptr(1..2, 1),
+                    Op::inc(1..2, 1),
+                    Op::dec_ptr(1..2, 1),
+                    Op::put_char(1..2),
+                ], 1, LoopDecrement::Pre, BlockInfo::new_empty()),
+            ]
+        };
+
+        let input = b"";
+        let mut output = Vec::new();
+
+        let heap = run(&program, Cursor::new(input), &mut output);
+
+        assert_eq!(heap[0], 0);
+        assert_eq!(heap[1], 5);
+        assert_eq!(&output, b"\x04\x03\x02\x01\\0x0");
+    }
+
+    #[test]
+    fn test_i_loop_post() {
+        let program = Program {
+            ops: vec![
+                Op::set(0..1, 5),
+                Op::i_loop_with_decrement(1..4, vec![
+                    Op::put_char(1..2),
+                ], 1, LoopDecrement::Post, BlockInfo::new_empty()),
+            ]
+        };
+
+        let input = b"";
+        let mut output = Vec::new();
+
+        let heap = run(&program, Cursor::new(input), &mut output);
+
+        assert_eq!(heap[0], 0);
+        assert_eq!(&output, &[5, 4, 3, 2, 1]);
+    }
+
     #[test]
     fn test_c_loop() {
         let program = Program {
             ops: vec![
                 Op::set(0..1, 10),
                 Op::c_loop(1..4, vec![
-                    Op::inc(1..2, 1),
-                    Op::inc(1..2, 1),
-                ], 2, 5),
+                    Op::inc_with_offset(1..2, 2, 1),
+                    Op::inc_with_offset(1..2, 2, 1),
+                ], 5, BlockInfo::new_empty()),
             ]
         };
 
@@ -1342,14 +1485,56 @@ mod tests {
         assert_eq!(heap[2], 10);
     }
 
+    #[test]
+    fn test_c_loop_pre() {
+        let program = Program {
+            ops: vec![
+                Op::c_loop_with_decrement(1..4, vec![
+                    Op::inc_ptr(1..2, 1),
+                    Op::inc(1..2, 1),
+                    Op::dec_ptr(1..2, 1),
+                    Op::put_char(1..2),
+                ], 5, LoopDecrement::Pre, BlockInfo::new_empty()),
+            ]
+        };
+
+        let input = b"";
+        let mut output = Vec::new();
+
+        let heap = run(&program, Cursor::new(input), &mut output);
+
+        assert_eq!(heap[0], 0);
+        assert_eq!(heap[1], 5);
+        assert_eq!(&output, b"\x04\x03\x02\x01\\0x0");
+    }
+
+    #[test]
+    fn test_c_loop_post() {
+        let program = Program {
+            ops: vec![
+                Op::c_loop_with_decrement(1..4, vec![
+                    Op::put_char(1..2),
+                ], 5, LoopDecrement::Post, BlockInfo::new_empty()),
+            ]
+        };
+
+        let input = b"";
+        let mut output = Vec::new();
+
+        let heap = run(&program, Cursor::new(input), &mut output);
+
+        assert_eq!(heap[0], 0);
+        assert_eq!(&output, &[5, 4, 3, 2, 1]);
+    }
+
     #[test]
     fn test_tnz() {
         let program = Program {
             ops: vec![
                 Op::set(0..1, 1),
                 Op::t_nz(1..4, vec![
-                    Op::set(1..2, 10),
-                ], 2),
+                    Op::set_with_offset(1..2, 2, 10),
+                ], BlockInfo::new_empty()),
             ]
         };
 
@@ -1368,8 +1553,8 @@ mod tests {
         let program = Program {
             ops: vec![
                 Op::t_nz(1..4, vec![
-                    Op::set(1..2, 10),
-                ], 2),
+                    Op::set_with_offset(1..2, 2, 10),
+                ], BlockInfo::new_empty()),
             ]
         };
 
@@ -1409,7 +1594,7 @@ mod tests {
     #[test]
     fn test_test_io() {
         let mut program = parse(",[.,]").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
 
         let input = b"0123456789aZ";
         let mut output = Vec::new();
@@ -1423,7 +1608,8 @@ mod tests {
     #[test]
     fn test_cell_size() {
         let mut program = parse(include_str!("../../../test_programs/cell_size.bf")).unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -1432,4 +1618,25 @@ mod tests {
 
         assert_eq!(output, b"8 bit cells\n");
     }
+
+    /*
+     * Disabled because JIT compilation is too slow in debug mode
+    #[test]
+    fn test_awib() {
+        let mut program = parse(include_str!("../../../test_programs/awib.bf")).unwrap();
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
+        let input = include_bytes!("../../../test_programs/mandelbrot.bf");
+        let mut output = Vec::new();
+
+        let _heap = run(&program, Cursor::new(input), &mut output);
+
+        let o = String::from_utf8_lossy(&output);
+
+        // Change needed because result starts wth 0 byte which is translated to \0x0 as of now
+        let o2 = format!("\\0x0{}", &include_str!("../../../test_programs/mandelbrot.c")[1..]);
+
+        assert_eq!(o, o2);
+    }
+     */
 }
\ No newline at end of file
