@@ -1,9 +1,10 @@
-use std::io::{Read, ErrorKind};
+use std::io::{ErrorKind, Read};
 use std::io::Write;
+use std::ops::Range;
 
 use crate::errors::RuntimeError;
-use crate::parser::{Op, Program, OpType};
-use std::ops::Range;
+use crate::ir::ops::{Op, OpType, LoopDecrement};
+use crate::parser::Program;
 
 const MAX_HEAP_SIZE: usize = 16 * 1024 * 1024;
 
@@ -43,6 +44,9 @@ impl<R: Read, W: Write> Interpreter<R, W> {
 
     fn execute_op(&mut self, op: &Op) -> Result<(), RuntimeError> {
         match &op.op_type {
+            OpType::Start => {
+                // ignore
+            }
             OpType::IncPtr(count) => self.pointer = self.pointer.wrapping_add(*count),
             OpType::DecPtr(count) => self.pointer = self.pointer.wrapping_sub(*count),
             OpType::Inc(offset, count) => {
@@ -94,6 +98,17 @@ impl<R: Read, W: Write> Interpreter<R, W> {
                 let target = self.heap_value_at_offset(&op.span, *dest_offset)?;
                 *target = target.wrapping_sub(*value);
             }
+            OpType::Mul(src_offset, dest_offset, multi) => {
+                let source = *self.heap_value_at_offset(&op.span, *src_offset)?;
+                let target = self.heap_value_at_offset(&op.span, *dest_offset)?;
+                *target = source.wrapping_mul(*multi);
+                *self.heap_value_at_offset(&op.span, *src_offset)? = 0;
+            }
+            OpType::NzMul(src_offset, dest_offset, multi) => {
+                let source = *self.heap_value_at_offset(&op.span, *src_offset)?;
+                let target = self.heap_value_at_offset(&op.span, *dest_offset)?;
+                *target = source.wrapping_mul(*multi);
+            }
             OpType::Move(src_offset, dest_offset) => {
                 let source = *self.heap_value_at_offset(&op.span, *src_offset)?;
                 let target = self.heap_value_at_offset(&op.span, *dest_offset)?;
@@ -105,56 +120,108 @@ impl<R: Read, W: Write> Interpreter<R, W> {
                 let target = self.heap_value_at_offset(&op.span, *dest_offset)?;
                 *target = source;
             }
-            OpType::GetChar => self.get_char(&op.span)?,
-            OpType::PutChar => self.put_char(&op.span)?,
-            OpType::DLoop(ops) => {
+            OpType::GetChar(offset) => self.get_char(&op.span, *offset)?,
+            OpType::PutChar(offset) => self.put_char(&op.span, *offset)?,
+            OpType::PutString(array) => self.put_string(&op.span, array)?,
+            OpType::DLoop(ops, _) => {
                 while *self.heap_value(&op.span)? > 0 {
                     self.execute_ops(ops)?;
                 }
             }
-            OpType::LLoop(ops, offset) => {
+            OpType::LLoop(ops, _) => {
                 let heap_pointer = self.pointer;
-                let start_heap_pointer = (heap_pointer as isize).wrapping_add(*offset) as usize;
 
                 while *self.heap_value(&op.span)? > 0 {
-                    self.pointer = start_heap_pointer;
                     self.execute_ops(ops)?;
                     self.pointer = heap_pointer;
                 }
             }
-            OpType::ILoop(ops, offset, step) => {
-                let heap_pointer = self.pointer;
-                let start_heap_pointer = (heap_pointer as isize).wrapping_add(*offset) as usize;
+            OpType::ILoop(ops, step, decrement, _) => {
+                match decrement {
+                    LoopDecrement::Pre => {
+                        let heap_pointer = self.pointer;
 
-                let mut left = *self.heap_value(&op.span)?;
-                while left > 0 {
-                    self.pointer = start_heap_pointer;
+                        let mut left = *self.heap_value(&op.span)?;
+                        while left > 0 {
+                            left = left.wrapping_sub(*step);
+                            *self.heap_value(&op.span)? = left;
+                            self.execute_ops(ops)?;
+                            self.pointer = heap_pointer;
+                        }
 
-                    self.execute_ops(ops)?;
-                    left = left.wrapping_sub(*step);
+                        *self.heap_value(&op.span)? = 0;
+                    }
+                    LoopDecrement::Post => {
+                        let heap_pointer = self.pointer;
+
+                        let mut left = *self.heap_value(&op.span)?;
+                        while left > 0 {
+                            self.execute_ops(ops)?;
+                            left = left.wrapping_sub(*step);
+                            *self.heap_value(&op.span)? = left;
+                            self.pointer = heap_pointer;
+                        }
+
+                        *self.heap_value(&op.span)? = 0;
+                    }
+                    LoopDecrement::Auto => {
+                        let heap_pointer = self.pointer;
+
+                        let mut left = *self.heap_value(&op.span)?;
+                        while left > 0 {
+                            self.execute_ops(ops)?;
+                            left = left.wrapping_sub(*step);
+                            self.pointer = heap_pointer;
+                        }
+
+                        *self.heap_value(&op.span)? = 0;
+                    }
                 }
-
-                self.pointer = heap_pointer;
-                *self.heap_value(&op.span)? = 0;
             }
-            OpType::CLoop(ops, offset, iterations) => {
-                let heap_pointer = self.pointer;
-                let start_heap_pointer = (heap_pointer as isize).wrapping_add(*offset) as usize;
+            OpType::CLoop(ops, iterations, decrement, _) => {
+                match decrement {
+                    LoopDecrement::Pre => {
+                        let heap_pointer = self.pointer;
 
-                for _ in 0..*iterations {
-                    self.pointer = start_heap_pointer;
-                    self.execute_ops(ops)?;
+                        *self.heap_value(&op.span)? = *iterations;
+                        let mut left = *self.heap_value(&op.span)?;
+                        while left > 0 {
+                            left = left.wrapping_sub(1);
+                            *self.heap_value(&op.span)? = left;
+                            self.execute_ops(ops)?;
+                            self.pointer = heap_pointer;
+                        }
+
+                        *self.heap_value(&op.span)? = 0;
+                    }
+                    LoopDecrement::Post => {
+                        let heap_pointer = self.pointer;
+
+                        *self.heap_value(&op.span)? = *iterations;
+                        let mut left = *self.heap_value(&op.span)?;
+                        while left > 0 {
+                            self.execute_ops(ops)?;
+                            self.pointer = heap_pointer;
+                            left = left.wrapping_sub(1);
+                            *self.heap_value(&op.span)? = left;
+                        }
+
+                        *self.heap_value(&op.span)? = 0;
+                    }
+                    LoopDecrement::Auto => {
+                        let heap_pointer = self.pointer;
+                        for _ in 0..*iterations {
+                            self.execute_ops(ops)?;
+                            self.pointer = heap_pointer;
+                        }
+                        *self.heap_value(&op.span)? = 0;
+                    }
                 }
-
-                self.pointer = heap_pointer;
-                *self.heap_value(&op.span)? = 0;
             }
-            OpType::TNz(ops, offset) => {
+            OpType::TNz(ops, _) => {
                 if *self.heap_value(&op.span)? != 0 {
                     let heap_pointer = self.pointer;
-                    let start_heap_pointer = (heap_pointer as isize).wrapping_add(*offset) as usize;
 
-                    self.pointer = start_heap_pointer;
                     self.execute_ops(ops)?;
 
                     self.pointer = heap_pointer;
@@ -232,7 +299,7 @@ impl<R: Read, W: Write> Interpreter<R, W> {
         Ok(&mut self.heap[pointer])
     }
 
-    fn get_char(&mut self, span: &Range<usize>) -> Result<(), RuntimeError> {
+    fn get_char(&mut self, span: &Range<usize>, offset: isize) -> Result<(), RuntimeError> {
         let mut buf = [0];
 
         if let Err(error) = self.input.read_exact(&mut buf) {
@@ -245,13 +312,13 @@ impl<R: Read, W: Write> Interpreter<R, W> {
             }
         };
 
-        *self.heap_value(span)? = buf[0];
+        *self.heap_value_at_offset(span, offset)? = buf[0];
 
         Ok(())
     }
 
-    fn put_char(&mut self, span: &Range<usize>) -> Result<(), RuntimeError> {
-        let ch = *self.heap_value(span)?;
+    fn put_char(&mut self, span: &Range<usize>, offset: isize) -> Result<(), RuntimeError> {
+        let ch = *self.heap_value_at_offset(span, offset)?;
 
         if ch.is_ascii() {
             write!(self.output, "{}", ch as char)
@@ -262,6 +329,21 @@ impl<R: Read, W: Write> Interpreter<R, W> {
             error,
         })
     }
+
+    fn put_string(&mut self, span: &Range<usize>, array: &[u8]) -> Result<(), RuntimeError> {
+        for &ch in array {
+            if ch.is_ascii() {
+                write!(self.output, "{}", ch as char)
+            } else {
+                write!(self.output, "\\0x{:x}", ch)
+            }.map_err(|error| RuntimeError::IoError {
+                span: span.clone(),
+                error,
+            })?;
+        }
+
+        Ok(())
+    }
 }
 
 
@@ -271,12 +353,15 @@ mod tests {
 
     use crate::backends::interpreter::Interpreter;
     use crate::parser::parse;
-    use crate::optimizations::optimize;
+    use crate::{Program, optimize_with_config, OptimizeConfig};
+    use crate::ir::ops::{Op, LoopDecrement};
+    use crate::ir::opt_info::BlockInfo;
 
     #[test]
     fn test_out_1() {
         let mut program = parse(">+.").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -289,7 +374,7 @@ mod tests {
     #[test]
     fn test_out_b() {
         let mut program = parse(",++.").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
 
         let input = b"a";
         let mut output = Vec::new();
@@ -302,7 +387,8 @@ mod tests {
     #[test]
     fn test_loop() {
         let mut program = parse("+++[>+<-]>.").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -319,7 +405,8 @@ mod tests {
                 >++++++++[<++++>-] <.>+++++++++++[<++++++++>-]<-.--------.+++
                 .------.--------.[-]>++++++++[<++++>- ]<+.[-]++++++++++."
         ).unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -332,7 +419,8 @@ mod tests {
     #[test]
     fn test_hello_world_v2() {
         let mut program = parse(include_str!("../../../test_programs/hello_world.bf")).unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -345,7 +433,8 @@ mod tests {
     #[test]
     fn test_count_loop() {
         let mut program = parse("++++[->++<]").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -361,7 +450,8 @@ mod tests {
     #[test]
     fn test_count_loop_inv() {
         let mut program = parse("++++[->++<]").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -377,7 +467,8 @@ mod tests {
     #[test]
     fn test_totally_empty() {
         let mut program = parse("").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -392,7 +483,8 @@ mod tests {
     #[test]
     fn test_optimized_empty() {
         let mut program = parse("[>+++++<-]").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -407,7 +499,8 @@ mod tests {
     #[test]
     fn test_multiply() {
         let mut program = parse(">>++<<++[>+++++<-]").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -423,7 +516,8 @@ mod tests {
     #[test]
     fn test_divide() {
         let mut program = parse(">>++<<++++++++++++[>+<---]").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -439,7 +533,8 @@ mod tests {
     #[test]
     fn test_pow_5_3() {
         let mut program = parse("+++++[>+++++[>+++++<-]<-]").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -456,7 +551,8 @@ mod tests {
     #[test]
     fn test_bad_count_loop() {
         let mut program = parse("++++++++[---->+<++]").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -472,7 +568,8 @@ mod tests {
     #[test]
     fn test_conditional_set() {
         let mut program = parse("+>>++<<[->[-]++++++<]").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -489,7 +586,8 @@ mod tests {
     #[test]
     fn test_overwrite_prev_set() {
         let mut program = parse("++++[-]++").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -504,7 +602,8 @@ mod tests {
     #[test]
     fn test_hello_fizz() {
         let mut program = parse(include_str!("../../../test_programs/fizz.bf")).unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -517,7 +616,8 @@ mod tests {
     #[test]
     fn test_hello_fizzbuzz() {
         let mut program = parse(include_str!("../../../test_programs/fizzbuzz.bf")).unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -530,7 +630,8 @@ mod tests {
     #[test]
     fn test_bottles() {
         let mut program = parse(include_str!("../../../test_programs/bottles.bf")).unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -543,7 +644,8 @@ mod tests {
     #[test]
     fn test_factor() {
         let mut program = parse(include_str!("../../../test_programs/factor.bf")).unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = include_bytes!("../../../test_programs/factor.bf.in");
         let mut output = Vec::new();
@@ -556,7 +658,8 @@ mod tests {
     #[test]
     fn test_life() {
         let mut program = parse(include_str!("../../../test_programs/life.bf")).unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = include_bytes!("../../../test_programs/life.bf.in");
         let mut output = Vec::new();
@@ -569,7 +672,8 @@ mod tests {
     #[test]
     fn test_test_io() {
         let mut program = parse(",[.,]").unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"0123456789aZ";
         let mut output = Vec::new();
@@ -582,7 +686,8 @@ mod tests {
     #[test]
     fn test_cell_size() {
         let mut program = parse(include_str!("../../../test_programs/cell_size.bf")).unwrap();
-        optimize(&mut program);
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
 
         let input = b"";
         let mut output = Vec::new();
@@ -591,4 +696,107 @@ mod tests {
 
         assert_eq!(output, b"8 bit cells\n");
     }
+
+    #[test]
+    fn test_awib() {
+        let mut program = parse(include_str!("../../../test_programs/awib.bf")).unwrap();
+        optimize_with_config(&mut program, &OptimizeConfig::o3());
+
+        let input = include_bytes!("../../../test_programs/mandelbrot.bf");
+        let mut output = Vec::new();
+
+        Interpreter::new(Cursor::new(input), &mut output).execute(&program).unwrap();
+
+        assert_eq!(output, include_bytes!("../../../test_programs/mandelbrot.c"));
+    }
+
+    #[test]
+    fn test_i_loop_pre() {
+        let program = Program {
+            ops: vec![
+                Op::set(0..1, 5),
+                Op::i_loop_with_decrement(1..4, vec![
+                    Op::inc_ptr(1..2, 1),
+                    Op::inc(1..2, 1),
+                    Op::dec_ptr(1..2, 1),
+                    Op::put_char(1..2),
+                ], 1, LoopDecrement::Pre, BlockInfo::new_empty()),
+            ]
+        };
+
+        let input = b"";
+        let mut output = Vec::new();
+
+        let mut interpreter = Interpreter::new(Cursor::new(input), &mut output);
+        interpreter.execute(&program).unwrap();
+
+        assert_eq!(interpreter.heap[0], 0);
+        assert_eq!(interpreter.heap[1], 5);
+        assert_eq!(&output, &[4, 3, 2, 1, 0]);
+    }
+
+    #[test]
+    fn test_i_loop_post() {
+        let program = Program {
+            ops: vec![
+                Op::set(0..1, 5),
+                Op::i_loop_with_decrement(1..4, vec![
+                    Op::put_char(1..2),
+                ], 1, LoopDecrement::Post, BlockInfo::new_empty()),
+            ]
+        };
+
+        let input = b"";
+        let mut output = Vec::new();
+
+        let mut interpreter = Interpreter::new(Cursor::new(input), &mut output);
+        interpreter.execute(&program).unwrap();
+
+        assert_eq!(interpreter.heap[0], 0);
+        assert_eq!(&output, &[5, 4, 3, 2, 1]);
+    }
+
+    #[test]
+    fn test_c_loop_pre() {
+        let program = Program {
+            ops: vec![
+                Op::c_loop_with_decrement(1..4, vec![
+                    Op::inc_ptr(1..2, 1),
+                    Op::inc(1..2, 1),
+                    Op::dec_ptr(1..2, 1),
+                    Op::put_char(1..2),
+                ], 5, LoopDecrement::Pre, BlockInfo::new_empty()),
+            ]
+        };
+
+        let input = b"";
+        let mut output = Vec::new();
+
+        let mut interpreter = Interpreter::new(Cursor::new(input), &mut output);
+        interpreter.execute(&program).unwrap();
+
+        assert_eq!(interpreter.heap[0], 0);
+        assert_eq!(interpreter.heap[1], 5);
+        assert_eq!(&output, &[4, 3, 2, 1, 0]);
+    }
+
+    #[test]
+    fn test_c_loop_post() {
+        let program = Program {
+            ops: vec![
+                Op::c_loop_with_decrement(1..4, vec![
+                    Op::put_char(1..2),
+                ], 5, LoopDecrement::Post, BlockInfo::new_empty()),
+            ]
+        };
+
+        let input = b"";
+        let mut output = Vec::new();
+
+        let mut interpreter = Interpreter::new(Cursor::new(input), &mut output);
+        interpreter.execute(&program).unwrap();
+
+        assert_eq!(interpreter.heap[0], 0);
+        assert_eq!(&output, &[5, 4, 3, 2, 1]);
+    }
 }
